//! Program for merging some sorted files with large sizes.
/*! Use a Forecasting merge algorithm, in the style
   suggested by Knuth volume 3 (2nd edition), exercise 5.4.6F.  Use
   the optimization suggested by exercise 5.4.9.
*/

#define _FILE_OFFSET_BITS 64
#undef ENABLE_NLS

#include <limits.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <error.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <fcntl.h>
#include <pthread.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>

#if HAVE_SYS_RESOURCE_H
# include <sys/resource.h>
#endif

#ifndef RLIMIT_DATA
struct rlimit
{
    size_t rlim_cur;
};
# define getrlimit(Resource, Rlp) (-1)
#endif

#define program_name "for_merge"

/*! Exit statuses.  */
enum
{
    /*! POSIX says to exit with status 1 if invoked with -c and the input is not properly sorted.  */
    SORT_OUT_OF_ORDER = 1,

    /*! POSIX says any other irregular exit must exit with a status code greater than 1.  */
    MERGE_FAILURE = 2
};

#define NONZERO(x) ((x) != 0)

/*! The character marking end of line. Default to \n. */
static char eolchar = '\n';

/*! Lines are held in core as counted strings. */
struct line
{
    char *text;                 /*!< Text of the line. */
    size_t length;              /*!< Length including final newline. */
};

//! Input buffers.
/*! | | | | | | | | | | | | | | | | | | ||||||||||||||
 *  |                                   |            |
 *  |      buffer's data                | lines data |
 *  |                                   |            |
 *  |0 1 2 3 4 ...                      |   ... 43210|
 *  ^                                   ^        ^   ^
 *  |                                   |        |   |
 *  buf                                base     cur  linelim
 */
struct buffer
{
    char *buf;                  /*!< Dynamically allocated buffer,
                                   partitioned into 3 regions:
                                   - input data;
                                   - unused area;
                                   - an array of lines, in reverse order.  */
    size_t used;                /*!< Number of bytes used for input data.  */
    size_t nlines;              /*!< Number of lines in the line array.  */
    size_t alloc;               /*!< Number of bytes allocated. */
    size_t left;                /*!< Number of bytes left from previous reads. */
    size_t line_bytes;          /*!< Number of bytes to reserve for each line. */
    bool eof;                   /*!< An EOF has been read.  */
};

#ifndef MAX
# define MAX(a, b) ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
# define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

/*! Minimum size for a buffer.  */
#define MIN_MERGE_BUFFER_SIZE (2 + sizeof (struct line))

/*! Minimum size for a merge.  */
#define MIN_MERGE_SIZE (16 * MIN_MERGE_BUFFER_SIZE)

/*! The number of bytes needed for a merge.  */
static size_t merge_buffer_size = MAX (MIN_MERGE_BUFFER_SIZE, 256 * 1024);

/*! The number of bytes needed for output buffer.  */
static size_t out_buffer_size = MAX (MIN_MERGE_BUFFER_SIZE, 256 * 1024);

/*! Some systems do not define EXIT_*, despite otherwise supporting C89.  */
# define EXIT_SUCCESS 0
# define EXIT_FAILURE 1

/*! This flag says that input buffer was filled and ready to use. */
bool flag_fcst_buffer_ready;
pthread_cond_t flag_buffer_ready_cv;
pthread_mutex_t flag_buffer_ready_mutex;

/*! This flag wakes up thread to fill input forecasted buffer from disk. */
bool flag_fcst_buffer_start;
pthread_cond_t flag_fillbuf_start_cv;
pthread_mutex_t flag_fillbuf_start_mutex;
struct buffer *fillbuf_buf;
struct buffer *fillbuf_buf2;
FILE *fillbuf_fp;
char const *fillbuf_file;
bool fillbuf_ret;

/*! Report MESSAGE for FILE, then clean up and exit. If FILE is null, it represents standard output.  */
static void
die (char const *message, char const *file)
{
    error (MERGE_FAILURE, errno, "%s: %s", message, file ? file : ("standard output"));
}

/*! Allocate N bytes of memory dynamically, with error checking.  */
void *
xmalloc (size_t n, size_t s)
{
    size_t ns = n * s;
    void *p = malloc (ns);
    if (!p && ns != 0)
        error (MERGE_FAILURE, 0, "memory exhausted");
    return p;
}

void
usage (int status)
{
    if (status != EXIT_SUCCESS)
        fprintf (stderr, "Try `%s --help' for more information.\n", program_name);
    else
    {
        printf (("\
Usage: %s [OPTION]... [FILE1]...\n\
"), program_name);
        fputs (("\
Write concatenation of sorted FILEs to file.\n\
\n\
"), stdout);
        fputs (("\
Mandatory arguments to long options are mandatory for short options too.\n\
"), stdout);
        fputs (("\
Options:\n\
\n\
"), stdout);
        fputs (("\
  -o, --output=OFILE        MANDATORY - write result to OFILE\n\
  -s, --buffer-size=SIZE    use SIZE bytes for input buffer (2*SIZE for each input file)\n\
  -S, --output-size=SIZE    use SIZE bytes for output buffer\n\
  -z, --zero-terminated     end lines with 0 byte, not newline\n\
  -h, --help                print this help\n\
"), stdout);
        fputs (("\
\n\
\n\
*** WARNING ***\n\
The locale specified by the environment doesn't affect sort order.\n\
LC_ALL=C is set to get the traditional sort order that uses\n\
native byte values.\n\
*** NOTE ***\n\
Forecasting merge algorithm (P-way merge) \n\
is required 2*P input buffers, where P - number of sorted input files.\n\
"), stdout);
    }

    exit (status);
}

static char const short_options[] = "o:s:S:zh";

static struct option const long_options[] = {
    {"output", required_argument, NULL, 'o'},
    {"buffer-size", required_argument, NULL, 's'},
    {"output-size", required_argument, NULL, 'S'},
    {"zero-terminated", no_argument, NULL, 'z'},
    {"help", no_argument, NULL, 'h'},
    {NULL, 0, NULL, 0},
};

struct mergefile
{
    char const *name;
};

static void
write_bytes (const char *buf, size_t n_bytes, FILE * fp, const char *output_file)
{
    if (fwrite (buf, 1, n_bytes, fp) != n_bytes)
        die (("write failed"), output_file);
}

/*! Return the total amount of physical memory.  */
double
physmem_total ()
{
#if defined _SC_PHYS_PAGES && defined _SC_PAGESIZE
    {                           /* This works on linux-gnu, solaris2 and cygwin.  */
        double pages = sysconf (_SC_PHYS_PAGES);
        double pagesize = sysconf (_SC_PAGESIZE);
        if (0 <= pages && 0 <= pagesize)
            return pages * pagesize;
    }
#endif
    return 0;
}


/*! Return the amount of physical memory available.  */
double
physmem_available ()
{
#if defined _SC_AVPHYS_PAGES && defined _SC_PAGESIZE
    {                           /* This works on linux-gnu, solaris2 and cygwin.  */
        double pages = sysconf (_SC_AVPHYS_PAGES);
        double pagesize = sysconf (_SC_PAGESIZE);
        if (0 <= pages && 0 <= pagesize)
            return pages * pagesize;
    }
#endif
    return 0;

}

/*! Return size of the available memory.  */
static size_t
get_memory_available (void)
{

    /* Let MEM be available memory or 1/8 of total memory, whichever
       is greater.  */
    double avail = physmem_available ();
    double total = physmem_total ();
    double mem = MAX (avail, total / 8);
    struct rlimit rlimit;

    /* Let SIZE be MEM, but no more than the maximum object size or
       system resource limits.  Avoid the MIN macro here, as it is not
       quite right when only one argument is floating point.  Don't
       bother to check for values like RLIM_INFINITY since in practice
       they are not much less than SIZE_MAX.  */
    size_t size = SIZE_MAX;
    if (mem < size)
        size = mem;
    if (getrlimit (RLIMIT_DATA, &rlimit) == 0 && rlimit.rlim_cur < size)
        size = rlimit.rlim_cur;
#ifdef RLIMIT_AS
    if (getrlimit (RLIMIT_AS, &rlimit) == 0 && rlimit.rlim_cur < size)
        size = rlimit.rlim_cur;
#endif

    /* Leave a large safety margin for the above limits, as failure can occur when they are exceeded.  */
    size /= 2;

#ifdef RLIMIT_RSS
    /* Leave a 1/16 margin for RSS to leave room for code, stack, etc.
       Exceeding RSS is not fatal, but can be quite slow.  */
    if (getrlimit (RLIMIT_RSS, &rlimit) == 0 && rlimit.rlim_cur / 16 * 15 < size)
        size = rlimit.rlim_cur / 16 * 15;
#endif

    /* Use no less than the minimum.  */
    return MAX (size, MIN_MERGE_SIZE);
}

/*! Initialize BUF.  Reserve LINE_BYTES bytes for each line; LINE_BYTES
   must be at least sizeof (struct line).  Allocate ALLOC bytes initially. */

static void
initbuf (struct buffer *buf, size_t line_bytes, size_t alloc)
{
    /* Ensure that the line array is properly aligned. */
    alloc += sizeof (struct line) - alloc % sizeof (struct line);
    buf->buf = malloc (alloc);
    if (!buf->buf)
        error (MERGE_FAILURE, 0, "-memory exhausted");

    buf->line_bytes = line_bytes;
    buf->alloc = alloc;
    buf->used = buf->left = buf->nlines = 0;
    buf->eof = false;
}

/*! Return one past the limit of the line array.  */

static inline struct line *
buffer_linelim (struct buffer const *buf)
{
    return (struct line *) (buf->buf + buf->alloc);
}


/*! Fill forecast buffer in background. */
void *
fill_input_buffer_thread (void *args)
{
    char eol = eolchar;

    while (1)
    {

        pthread_mutex_lock (&flag_fillbuf_start_mutex);
        while (!flag_fcst_buffer_start)
            pthread_cond_wait (&flag_fillbuf_start_cv, &flag_fillbuf_start_mutex);

        struct buffer *buf = fillbuf_buf;
        struct buffer *buf2 = fillbuf_buf2;
        FILE *fp = fillbuf_fp;
        char const *file = fillbuf_file;
        flag_fcst_buffer_start = false;
        fillbuf_ret = false;
        pthread_mutex_unlock(&flag_fillbuf_start_mutex);

        size_t line_bytes = buf->line_bytes;

        if (buf->eof)
            goto fillbuf_ret_false;

        //  check unprocessed symbols
        if (buf2->used != buf2->left)
        {
            memmove (buf->buf, buf2->buf + buf2->used - buf2->left, buf2->left);
            buf->used = buf2->left;
            buf->nlines = 0;
            buf->left = buf2->left;
        }

        for (;;)
        {
            char *ptr = buf->buf + buf->used;
            struct line *linelim = buffer_linelim (buf);
            struct line *line = linelim - buf->nlines;
            size_t avail = (char *) linelim - buf->nlines * line_bytes - ptr;
            char *line_start = buf->nlines ? line->text + line->length : buf->buf;

            while (line_bytes + 1 < avail)
            {
                /* Read as many bytes as possible, but do not read so many
                   bytes that there might not be enough room for the
                   corresponding line array. */
                size_t readsize = (avail - 1) / (line_bytes + 1);

                size_t bytes_read = fread (ptr, 1, readsize, fp);
                char *ptrlim = ptr + bytes_read;
                char *p;
                avail -= bytes_read;

                if (bytes_read != readsize)
                {
                    if (ferror (fp))
                        die (("read failed"), file);
                    if (feof (fp))
                    {
                        buf->eof = true;
                        if (buf->buf == ptrlim)
                            goto fillbuf_ret_false;
                        if (ptrlim[-1] != eol)
                            *ptrlim++ = eol;
                    }
                }

                /* Find and record each line in the just-read input.  */
                while ((p = memchr (ptr, eol, ptrlim - ptr)))
                {
                    ptr = p + 1;
                    line--;
                    line->text = line_start;
                    line->length = ptr - line_start;
                    avail -= line_bytes;

                    line_start = ptr;
                }

                ptr = ptrlim;
                if (buf->eof)
                    break;
            }

            buf->used = ptr - buf->buf;
            buf->nlines = buffer_linelim (buf) - line;
            if (buf->nlines != 0)
            {
                buf->left = ptr - line_start;
                goto fillbuf_ret_true;
            }
        }

      fillbuf_ret_true:

        fillbuf_ret = true;

      fillbuf_ret_false:

        pthread_mutex_lock (&flag_buffer_ready_mutex);
        flag_fcst_buffer_ready = true;
        pthread_cond_signal (&flag_buffer_ready_cv);
        pthread_mutex_unlock (&flag_buffer_ready_mutex);

    }                           // while (1)
}


/*! Compare two lines A and B, returning negative, zero, or positive
   depending on whether A compares less than, equal to, or greater than B. */

static int
compare (const struct line *a, const struct line *b)
{
    int diff;
    size_t alen, blen;

    /* If the keys all compare equal (or no keys were specified)
       fall through to the default comparison.  */
    alen = a->length - 1, blen = b->length - 1;

    if (alen == 0)
        diff = -NONZERO (blen);
    else if (blen == 0)
        diff = 1;
    else if (!(diff = memcmp (a->text, b->text, MIN (alen, blen))))
        diff = alen < blen ? -1 : alen != blen;

    return diff;
}

/*! The new line just read in may be larger than other lines already in main memory; push it back in the queue until we
    encounter a line larger than it.  Optimize for the common case where the new line is smallest.  */
inline void
reconstruct_table (size_t nfiles, struct line const **cur, size_t * ord)
{
    int j;
    size_t lo = 1;
    size_t hi = nfiles;
    size_t probe = lo;
    size_t ord0 = ord[0];
    size_t count_of_smaller_lines;

    while (lo < hi)
    {
        int cmp = compare (cur[ord0], cur[ord[probe]]);
        if (cmp < 0 || (cmp == 0 && ord0 < ord[probe]))
            hi = probe;
        else
            lo = probe + 1;
        probe = (lo + hi) / 2;
    }

    count_of_smaller_lines = lo - 1;
    for (j = 0; j < count_of_smaller_lines; j++)
        ord[j] = ord[j + 1];
    ord[count_of_smaller_lines] = ord0;
}

/*! Open FILES (there are NFILES of them) and store the resulting array of stream pointers into (*PFPS).  Allocate the array.  Return the
   number of successfully opened files, setting errno if this value is less than NFILES. */

static size_t
open_input_files (struct mergefile *files, size_t nfiles, FILE *** pfps, char *const *f)
{
    FILE **fps = *pfps = xmalloc (nfiles, sizeof *fps);
    int i;

    /* Open input files with mmap for reading. */
    for (i = 0; i < nfiles; i++)
    {
        fps[i] = fopen (files[i].name, "rm");
        if (!fps[i])
            break;
    }

    return i;
}

/*! Start filling input buffer in background thread.
    Use second buffer BUF2 to check whether it has some unprocessed data.
    If it has - copy this data to the begin of the first buffer. */
void
fill_input_buffer (struct buffer *buf, struct buffer *buf2, FILE * fp, char const *file)
{
    pthread_mutex_lock (&flag_fillbuf_start_mutex);
    fillbuf_buf = buf;
    fillbuf_buf2 = buf2;
    fillbuf_fp = fp;
    fillbuf_file = file;
    flag_fcst_buffer_start = true;
    pthread_cond_signal (&flag_fillbuf_start_cv);
    pthread_mutex_unlock (&flag_fillbuf_start_mutex);
}

/*! Swap buffers: first one has new data, second has old data. */
bool
swap_buffers (struct buffer *from, struct buffer *to)
{
    pthread_mutex_lock (&flag_buffer_ready_mutex);
    while (!flag_fcst_buffer_ready)
        pthread_cond_wait (&flag_buffer_ready_cv, &flag_buffer_ready_mutex);
    flag_fcst_buffer_ready = false;
    bool return_code = fillbuf_ret;

    to->line_bytes = from->line_bytes;
    to->alloc = from->alloc;
    to->used = from->used;
    to->left = from->left;
    to->nlines = from->nlines;
    to->eof = from->eof;
    char *tmp = to->buf;
    to->buf = from->buf;
    from->buf = tmp;

    pthread_mutex_unlock (&flag_buffer_ready_mutex);
    
    return return_code;
}

/*! Merge lines from FILES onto OFP.  NFILES is the number of files; FPS is the vector of open stream corresponding to the files.
   Close input and output streams before returning. OUTPUT_FILE gives the name of the output file. */

static void
merge_fps (struct mergefile *files, size_t nfiles, FILE * ofp, char const *output_file, FILE ** fps)
{
    struct buffer *buffer = (struct buffer *) xmalloc (nfiles, sizeof *buffer); /*!< Input buffers for each file. */
    struct line const **cur = (struct line const **) xmalloc (nfiles, sizeof *cur); /*!< Current line in each line table. */
    struct line const **base = (struct line const **) xmalloc (nfiles, sizeof *base); /*!< Base of each line table.  */
    size_t *ord = (size_t *) xmalloc (nfiles, sizeof *ord); /*!< Table representing a permutation of fps, such that cur[ord[0]] is the smallest line and will be next output. */
    struct buffer *buffer_fcst = (struct buffer *) xmalloc (nfiles, sizeof *buffer); /*!< Buffers for forecasting. */
    size_t *ord_base = (size_t *) xmalloc (nfiles, sizeof *ord_base); /*!< Table representing a permutation of buffers' ends, such that buffer_fcst[ord_base[0]] is the buffer with smallest line and will be emptied first (and should be filled first) */


    size_t i;
    size_t j;
    size_t t;

    /*! output buffer */
    char *out_buf = malloc (out_buffer_size);
    if (!out_buf)
        error (MERGE_FAILURE, 0, "memory exhausted");
    char *out_buf_cur = out_buf;
    size_t out_buf_len_cur = out_buffer_size;

    /* Prepare buffers and read initial lines from each input file. */
    for (i = 0; i < nfiles;)
    {
        initbuf (&buffer[i], sizeof (struct line), merge_buffer_size);
        initbuf (&buffer_fcst[i], sizeof (struct line), merge_buffer_size);

        fill_input_buffer (&buffer_fcst[i], &buffer[i], fps[i], files[i].name);

        if (swap_buffers (&buffer_fcst[i], &buffer[i]))
        {
            struct line const *linelim = buffer_linelim (&buffer[i]);
            cur[i] = linelim - 1;
            base[i] = linelim - buffer[i].nlines;
            i++;
        }
        else
        {
            // fps[i] is empty; eliminate it from future consideration.  
            if (fclose (fps[i]) != 0)
                die (("close failed"), files[i].name);
            free (buffer[i].buf);
            free (buffer_fcst[i].buf);
            --nfiles;
            for (j = i; j < nfiles; ++j)
            {
                files[j] = files[j + 1];
                fps[j] = fps[j + 1];
            }
        }

    }

    /* Set up the ord table according to comparisons among input lines.
       Since this only reorders two items if one is strictly greater than
       the other, it is stable. */
    for (i = 0; i < nfiles; ++i) {
        ord[i] = i;
    }
    for (i = 1; i < nfiles; ++i) {
        if (0 < compare (cur[ord[i - 1]], cur[ord[i]]))
            t = ord[i - 1], ord[i - 1] = ord[i], ord[i] = t, i = 0;
    }

    /* Set up the ord_base table according to comparisons among input lines. */
    for (i = 0; i < nfiles; ++i)
        ord_base[i] = i;
    for (i = 1; i < nfiles; ++i)
        if (0 < compare (base[ord_base[i - 1]], base[ord_base[i]]))
            t = ord_base[i - 1], ord_base[i - 1] = ord_base[i], ord_base[i] = t, i = 0;

    fill_input_buffer (&buffer_fcst[ord_base[0]], &buffer[ord_base[0]], fps[ord_base[0]], files[ord_base[0]].name);

    /* Repeatedly output the smallest line until no input remains. */
    while (nfiles)
    {
        struct line const *smallest = cur[ord[0]];

        if (out_buf_len_cur > smallest->length)
        {
            memmove (out_buf_cur, smallest->text, smallest->length);
            out_buf_len_cur -= smallest->length;
            out_buf_cur += smallest->length;
        }
        else
        {
            write_bytes (out_buf, out_buffer_size - out_buf_len_cur, ofp, output_file);
            out_buf_len_cur = out_buffer_size;
            out_buf_cur = out_buf;

            memmove (out_buf_cur, smallest->text, smallest->length);
            out_buf_len_cur -= smallest->length;
            out_buf_cur += smallest->length;
        }

        /* Check if we need to read more lines into core. */
        if (base[ord[0]] < smallest)
            cur[ord[0]] = smallest - 1;
        else
        {
            if (buffer[ord[0]].eof)
            {
                //We reached EOF on fps[ord[0]].  
                for (i = 1; i < nfiles; ++i)
                    if (ord[i] > ord[0])
                        --ord[i];

                for (i = 1; i < nfiles; ++i)
                    if (ord_base[i] > ord_base[0])
                        --ord_base[i];

                --nfiles;
                if (fclose (fps[ord[0]]) != 0)
                    die (("close failed"), files[ord[0]].name);

                free (buffer[ord[0]].buf);
                for (i = ord[0]; i < nfiles; ++i)
                {
                    fps[i] = fps[i + 1];
                    files[i] = files[i + 1];
                    buffer[i] = buffer[i + 1];
                    cur[i] = cur[i + 1];
                    base[i] = base[i + 1];
                }
                for (i = 0; i < nfiles; ++i)
                    ord[i] = ord[i + 1];

                free (buffer_fcst[ord_base[0]].buf);
                for (i = ord_base[0]; i < nfiles; ++i)
                {
                    buffer_fcst[i] = buffer_fcst[i + 1];
                }
                for (i = 0; i < nfiles; ++i)
                    ord_base[i] = ord_base[i + 1];

                continue;
            }
            else
            {
                /* The end of buffer has just been reached.
                   Forecasted buffer has already been filled and ready to use. */
                bool return_code = swap_buffers (&buffer_fcst[ord_base[0]], &buffer[ord[0]]);

                struct line const *linelim = buffer_linelim (&buffer[ord[0]]);
                cur[ord[0]] = linelim - 1;
                base[ord[0]] = linelim - buffer[ord[0]].nlines;

                reconstruct_table (nfiles, base, ord_base);

                if (return_code)
                    fill_input_buffer (&buffer_fcst[ord_base[0]], &buffer[ord_base[0]], fps[ord_base[0]], files[ord_base[0]].name);
            }
        }
        reconstruct_table (nfiles, cur, ord);
    }                           // while (nfiles)


    if (out_buffer_size != out_buf_len_cur)
    {
        write_bytes (out_buf, out_buffer_size - out_buf_len_cur, ofp, output_file);
    }

    if (fclose (ofp) != 0)
        die (("close failed"), output_file);
    free (fps);
    free (buffer);
    free (buffer_fcst);
    free (ord);
    free (ord_base);
    free (base);
    free (cur);
    free (out_buf);
}


int
main (int argc, char **argv)
{
    int c = 0;
    size_t nfiles = 0;
    char **ifiles;
    char const *output_file = NULL;
    size_t merge_size = 0;
    size_t out_size = 0;
    FILE **fps;

    ifiles = (char **) xmalloc (argc, sizeof *ifiles);

    for (;;)
    {
        int oi = -1;

        if (c == -1 || ((c = getopt_long (argc, argv, short_options, long_options, &oi)) == -1))
        {
            if (argc <= optind)
                break;
            ifiles[nfiles++] = argv[optind++];
        }
        else
            switch (c)
            {
            case 1:
                ifiles[nfiles++] = optarg;
                break;

            case 'o':
                if (output_file && !strcmp (output_file, optarg))
                    error (MERGE_FAILURE, 0, ("multiple output files specified"));
                output_file = optarg;
                break;

            case 's':
                {
                    char *endptr;
                    merge_size = strtol (optarg, &endptr, 10);
                    if ((errno == ERANGE && (merge_size == LONG_MAX || merge_size == LONG_MIN)) || (errno != 0 && merge_size == 0))
                    {
                        error (MERGE_FAILURE, 0, ("strtol"));
                    }
                    break;
                }

            case 'S':
                {
                    char *endptr;
                    out_size = strtol (optarg, &endptr, 10);
                    if ((errno == ERANGE && (out_size == LONG_MAX || out_size == LONG_MIN)) || (errno != 0 && out_size == 0))
                    {
                        error (MERGE_FAILURE, 0, ("strtol"));
                    }
                    break;
                }

            case 'z':
                eolchar = 0;
                break;

            case 'h':
                usage (EXIT_SUCCESS);
                break;

            default:
                usage (EXIT_SUCCESS);
            }
    }

    if (nfiles < 2)
        error (MERGE_FAILURE, 0, ("More input files are required for merging"));

    setenv ("LC_ALL", "C", 1);

    size_t memory_available = get_memory_available ();
    if (merge_size > 0)
    {
        //  the user has specified a size
        merge_size = MAX (merge_size, MIN_MERGE_SIZE);
        merge_buffer_size = merge_size;
    }
    else
    {
        merge_buffer_size = MAX (merge_buffer_size, memory_available / (2 * nfiles + 2));
    }

    if (out_size > 0)
    {
        //  the user has specified a size
        out_size = MAX (out_size, out_buffer_size);
        out_buffer_size = out_size;
    }
    else
    {
        out_buffer_size = MAX (out_buffer_size, memory_available / (2 * nfiles + 2));
    }

    fprintf (stdout, "        merge_size = %ld\n", (long int) merge_size);
    fprintf (stdout, "          out_size = %ld\n", (long int) out_size);
    fprintf (stdout, "  memory_available = %ld\n", (long int) memory_available);
    fprintf (stdout, " merge_buffer_size = %ld\n", (long int) merge_buffer_size);
    fprintf (stdout, "   out_buffer_size = %ld\n", (long int) out_buffer_size);


    flag_fcst_buffer_ready = false;
    pthread_cond_init (&flag_buffer_ready_cv, NULL);
    pthread_mutex_init (&flag_buffer_ready_mutex, NULL);

    flag_fcst_buffer_start = false;
    fillbuf_buf = NULL;
    fillbuf_buf2 = NULL;
    fillbuf_fp = NULL;
    fillbuf_file = NULL;
    fillbuf_ret = false;
    pthread_cond_init (&flag_fillbuf_start_cv, NULL);
    pthread_mutex_init (&flag_fillbuf_start_mutex, NULL);

    pthread_t fillbuf_thread_id;
    pthread_create (&fillbuf_thread_id, NULL, &fill_input_buffer_thread, NULL);


    struct mergefile *files = (struct mergefile *) calloc (nfiles, sizeof *files);
    if (!files)
        error (MERGE_FAILURE, 0, "memory exhausted");

    size_t i;
    for (i = 0; i < nfiles; ++i)
        files[i].name = ifiles[i];

    size_t nopened = open_input_files (files, nfiles, &fps, ifiles);

    if (nopened == nfiles)
    {
        FILE *ofp = fopen (output_file, "w");
        if (ofp)
        {
            merge_fps (files, nfiles, ofp, output_file, fps);
        }
        else if (errno != EMFILE || nopened <= 2)
            die (("open failed"), output_file);
    }
    else if (nopened <= 2)
        die (("open failed"), files[nopened].name);

    pthread_cancel (fillbuf_thread_id);
    pthread_join (fillbuf_thread_id, NULL);

    free (files);

    exit (EXIT_SUCCESS);
}
